# -*- coding: utf-8 -*-

from bencoder import bdecode, bdecode2
import os
import sys

TORRENT_PATH = os.path.join(
    os.path.abspath(os.path.dirname(__file__)),
    "debian-8.3.0-amd64-netinst.iso.torrent"
)


def test_decode2():
    decoded, length = bdecode2(b'6:WWWWWWi233e')
    assert decoded == b'WWWWWW'
    assert length == 8


def test_decode_str():
    assert bdecode(b'6:WWWWWW') == b"WWWWWW"


def test_decode_int():
    assert bdecode(b'i233e') == 233


def test_decode_large_int():
    assert bdecode(b'i1455189890e') == 1455189890
    assert bdecode(b'i25735241490e') == 25735241490

    MAX_SIZE = sys.maxsize + 1
    BENCODED_MAXSIZE = ('i%de' % MAX_SIZE).encode()
    assert bdecode(BENCODED_MAXSIZE) == MAX_SIZE


def test_decode_list():
    assert bdecode(b'l1:a1:bi3ee') == [b'a', b'b', 3]


def test_decode_dict():
    od = dict()
    od[b'ka'] = b'va'
    od[b'kb'] = 2
    assert bdecode(b'd2:ka2:va2:kbi2ee') == od


def test_ordered_dict():
    from bencoder import OrderedDict
    rv = bdecode(b'd2:ka2:va2:kbi2ee')
    assert isinstance(rv, OrderedDict)
    assert list(rv.keys()) == [b'ka', b'kb']
    assert list(bdecode(b'd2:kc2:va2:kei2ee').keys()) == [b'kc', b'ke']
    assert list(bdecode(b'd2:ke2:va2:kci2ee').keys()) == [b'ke', b'kc']


def test_encode_complex():
    od = dict()
    od[b'KeyA'] = [b'listitemA', {b'k': b'v'}, 3]
    od[b'KeyB'] = {b'k': b'v'}
    od[b'KeyC'] = 3
    od[b'KeyD'] = b'AString'
    expected_result = b'd4:KeyAl9:listitemAd1:k1:vei3ee4:KeyBd1:k1:ve4:KeyCi3e4:KeyD7:AStringe'
    assert bdecode(expected_result) == od


def test_decode_debian_torrent():
    with open(TORRENT_PATH, "rb") as f:
        torrent = bdecode(f.read())
    assert torrent[b'announce'] == b'http://bttracker.debian.org:6969/announce'
    assert torrent[b'comment'] == b'"Debian CD from cdimage.debian.org"'



# -*- coding: utf-8 -*-

from bencoder import bencode, bdecode
import os
import sys

TORRENT_PATH = os.path.join(
    os.path.abspath(os.path.dirname(__file__)),
    "debian-8.3.0-amd64-netinst.iso.torrent"
)


def test_encode_str():
    assert bencode("WWWWWW") == b'6:WWWWWW'


def test_encode_int():
    assert bencode(233) == b'i233e'


def test_encode_large_int():
    assert bencode(1455189890) == b'i1455189890e'
    assert bencode(25735241490) == b'i25735241490e'
    MAX_SIZE = sys.maxsize + 1
    BENCODED_MAXSIZE = ('i%de' % MAX_SIZE).encode()

    assert bencode(MAX_SIZE) == BENCODED_MAXSIZE


def test_encode_bytes():
    b = b"TheseAreSomeBytes"
    coded = bencode(b)
    l = str(len(b)).encode()
    assert coded == l + b':' + b


def test_encode_string():
    b = "TheseAreSomeString"
    coded = bencode(b)
    l = str(len(b))
    assert coded == (l + ':' + b).encode()


def test_encode_list():
    assert bencode(['a', 'b', 3]) == b'l1:a1:bi3ee'


def test_encode_tuple():
    assert bencode(('a', 'b', 3)) == b'l1:a1:bi3ee'


def test_encode_true():
    assert bencode(True) == bencode(1)


def test_encode_false():
    assert bencode(False) == bencode(0)


def test_encode_dict():
    od = dict()
    od['ka'] = 'va'
    od['kb'] = 2
    assert bencode(od) == b'd2:ka2:va2:kbi2ee'


def test_encode_dict_subclass():
    class AAA(dict):
        pass

    od = AAA()
    od['ka'] = 'va'
    od['kb'] = 2
    assert bencode(od) == b'd2:ka2:va2:kbi2ee'


def test_encode_complex():
    od = dict()
    od['KeyA'] = ['listitemA', {'k': 'v'}, 3]
    od['KeyB'] = {'k': 'v'}
    od['KeyC'] = 3
    od['KeyD'] = 'AString'
    expected_result = b'd4:KeyAl9:listitemAd1:k1:vei3ee4:KeyBd1:k1:ve4:KeyCi3e4:KeyD7:AStringe'
    assert bencode(od) == expected_result


def test_infohash():
    import hashlib
    with open(TORRENT_PATH, "rb") as f:
        torrent = bdecode(f.read())
    infohash = hashlib.sha1(bencode(torrent[b'info'])).hexdigest()
    assert infohash == "4194e473d6c49630e1c6247d6716076809bb96ae"



# cython: language_level=3

# The contents of this file are subject to the BitTorrent Open Source License
# Version 1.1 (the License).  You may not copy or use this file, in either
# source code or executable form, except in compliance with the License.  You
# may obtain a copy of the License at http://www.bittorrent.com/license/.
#
# Software distributed under the License is distributed on an AS IS basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
# for the specific language governing rights and limitations under the
# License.

# Based on https://github.com/karamanolev/bencode3/blob/master/bencode.py



from cpython.version cimport PY_MAJOR_VERSION, PY_MINOR_VERSION
END_CHAR = ord('e')
ARRAY_TYPECODE = 'b'

if PY_MAJOR_VERSION >= 3 and PY_MINOR_VERSION >=7:
    OrderedDict = dict
else:
    from collections import OrderedDict

class BTFailure(Exception):
    pass


def decode_int(bytes x, int f):
    f += 1
    cdef long new_f = x.index(b'e', f)
    n = int(x[f:new_f])
    if x[f] == b'-'[0]:
        if x[f + 1] == b'0'[0]:
            raise ValueError()
    elif x[f] == b'0'[0] and new_f != f + 1:
        raise ValueError()
    return n, new_f + 1


def decode_string(bytes x, int f):
    cdef long colon = x.index(b':', f)
    cdef long n = int(x[f:colon])
    if x[f] == b'0'[0] and colon != f + 1:
        raise ValueError()
    colon += 1
    return x[colon:colon + n], colon + n


def decode_list(bytes x, int f):
    r, f = [], f + 1
    while x[f] != END_CHAR:
        v, f = decode_func[x[f]](x, f)
        r.append(v)
    return r, f + 1


def decode_dict(bytes x, int f):
    r = OrderedDict()
    f += 1
    while x[f] != END_CHAR:
        k, f = decode_string(x, f)
        r[k], f = decode_func[x[f]](x, f)
    return r, f + 1


decode_func = dict()

for func, keys in [
    (decode_list, 'l'),
    (decode_dict, 'd'),
    (decode_int, 'i'),
    (decode_string, [str(x) for x in range(10)])
]:
    for key in keys:
        decode_func[ord(key)] = func


def bdecode2(bytes x):
    try:
        r, l = decode_func[x[0]](x, 0)
    except (IndexError, KeyError, ValueError):
        raise BTFailure("not a valid bencoded string")
    return r, l

def bdecode(bytes x):
    r, l = bdecode2(x)
    if l != len(x):
        raise BTFailure("invalid bencoded value (data after valid prefix)")
    return r

cdef encode(v, list r):
    tp = type(v)
    if tp in encode_func:
        return encode_func[tp](v, r)
    else:
        for tp, func in encode_func.items():
            if isinstance(v, tp):
                return func(v, r)
    raise BTFailure(
        "Can't encode {0}(Type: {1})".format(v, type(v))
    )


cdef encode_int(long x, list r):
    r.append(b'i')
    r.append(str(x).encode())
    r.append(b'e')


cdef encode_long(x, list r):
    r.append(b'i')
    r.append(str(x).encode())
    r.append(b'e')


cdef encode_bytes(x, list r):
    r.append(str(len(x)).encode())
    r.append(b':')
    r.append(x)


cdef encode_string(str x, list r):
    r.append(str(len(x)).encode())
    r.append(b':')
    r.append(x.encode())


cdef encode_list(x, list r):
    r.append(b'l')
    for i in x:
        encode(i, r)
    r.append(b'e')


cdef encode_dict(x, list r):
    r.append(b'd')
    item_list = list(x.items())
    item_list.sort()
    for k, v in item_list:
        if isinstance(k, str):
            k = k.encode()
        encode_bytes(k, r)
        encode(v, r)
    r.append(b'e')


encode_func = {
    int: encode_int,
    bool: encode_int,    
    long: encode_long,
    bytes: encode_bytes,
    str: encode_string,
    list: encode_list,
    tuple: encode_list,
    dict: encode_dict,
    OrderedDict: encode_dict,
}


def bencode(x):
    r = []
    encode(x, r)
    return b''.join(r)



import os.path
import sys
import os

from setuptools import setup
from setuptools.extension import Extension
from setuptools.command.test import test as TestCommand

pyx_path = 'bencoder.pyx'
c_path = 'bencoder.c'

if os.path.exists(c_path):
    # Remove C file to force Cython recompile.
    os.remove(c_path)

if os.environ.get("BENCODER_LINETRACE", "") == "1":
    from Cython.Compiler.Options import get_directive_defaults
    directive_defaults = get_directive_defaults()

    directive_defaults['linetrace'] = True
    directive_defaults['binding'] = True

from Cython.Build import cythonize
ext_modules = cythonize(Extension(
    "bencoder",
    [pyx_path],
    extra_compile_args=['-O3']
))


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import coverage
        cov = coverage.Coverage()
        cov.start()

        import pytest
        errno = pytest.main(self.pytest_args)

        cov.stop()
        cov.save()
        sys.exit(errno)


cmdclass = {'test': PyTest}



setup(
    name='bencoder.pyx',
    version='3.0.0',
    description='Yet another bencode implementation in Cython',
    long_description=open('README.rst', 'r').read(),
    author='whtsky',
    author_email='whtsky@gmail.com',
    url='https://github.com/whtsky/bencoder.pyx',
    license='BSDv3',
    platforms=['POSIX', 'Windows'],
    zip_safe=False,
    include_package_data=True,
    keywords=['bencoding', 'encode', 'decode', 'bittorrent', 'bencode', 'bencoder', 'cython'],
    cmdclass=cmdclass,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Other Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Cython',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
    ],
    ext_modules=ext_modules,
    install_requires=[],
    tests_require=['cython', 'pytest', 'coverage'],
)
